(*PROGRAM: SOLIDS
  AUTHOR : JONATHAN KRAIDIN
  PURPOSE: In this program we will define a solid and
           be able to rotate, translate, and scale the body*)

PROGRAM Solids;

USES
  Dos,Crt, Graph;

CONST MatSize = 3;
      NoVertsMax = 8; {Maximum allowable vertices}
      NoLinesMax = 12;{Maximum number of lines}

TYPE MatType = array[1..MatSize,1..MatSize] of real;
VAR
  NoVerts,NoLines : word;
  Gd, Gm          : integer;
  ch              : char;
  X,Y,Z,xp,yp,zp  : array[1..NoVertsMax] of real;
  L               : array[1..2,1..NoLinesMax] of word;

  P          : MatType;         {results stored in this matrix}
  x1,y1,z1   : integer;         {translation variables}
  sx,sy,sz   : integer;         {scaling variables}
  axis_o     : word;            {axis of real-time rotation}
  dist       : word;            {distance object is viewed}
  choice     : word;
{===============================================================}
{Draw using cartesian coordinates.  Origin at screen center
 (320,100)}

PROCEDURE Make_Line(x1,y1,x2,y2 : real);

CONST X_cent = 320;
      Y_cent = 100;
      Scal_f = 2.20;
VAR
      a1,b1,a2,b2 : integer;
BEGIN
  a1 := round(scal_f*x1 + x_cent);
  a2 := round(scal_f*x2 + x_cent);
  b1 := round(y_cent-y1);
  b2 := round(y_cent-y2);
  line(a1,b1,a2,b2);
END;{end Make_line}
{===============================================================}
{In this procedure we hard-code in the X,Y,Z vertices
 into their respective one-dimensional arrays.  In
 addition, we hard-code the line segments into
 the two-dimensional array L}

PROCEDURE Set_Up;
                     {The vertices and lines define a cube.}
VAR                  {The array L tells how we connect}
    c : word;        {the vertices.}
BEGIN                {The format is V = L[Point,Line#], where Point}
                     {is either endpoint one or two, and Line# is}
  for c := 1 to 4 do {the line number we are at (twelve to a cube).}
    x[c] := 1;       {By plugging in the endpoint and line number we}
  for c := 5 to 8 do {get the vertex number.}
    x[c] := -1;
  for c := 1 to 2 do
   begin
    y[c]   := 1;
    y[c+2] := -1;
    y[c+4] := 1;
    y[c+6] := -1;
   end;
  z[1] := 1;z[3]:= 1;z[5]:= 1;z[7]:=1;
  z[2] := -1;z[4]:=-1;z[6]:=-1;z[8]:=-1;
                             {Now define how lines are connected}
  L[1,1] := 1;L[2,1] := 2;L[1,2] := 2;L[2,2] := 4;L[1,3] := 4;
  L[2,3] := 3;L[1,4] := 3;L[2,4] := 1;L[1,5] := 5;L[2,5] := 6;
  L[1,6] := 6;L[2,6] := 8;L[1,7] := 8;L[2,7] := 7;L[1,8] := 7;
  L[2,8] := 5;L[1,9] := 1;L[2,9] := 5;L[1,10]:= 2;L[2,10]:= 6;
  L[1,11]:= 4;L[2,11]:= 8;L[1,12]:= 3;L[2,12]:= 7;
END; {end SetUp}
{===============================================================}
{This Procedure will perform matrix multiplication,
 multiplying A x B and returning the
 cross-product matrix in C.}

PROCEDURE Mult(A,B:MatType;Var C : MatType);
VAR                   {This algorithm performs multiplication}
    Row,Col,k : word; {yielding:  a1 b1 c1           q1 r1 s1      }
    sum       : real; {           a2 b2 c2     x     q2 r2 s2    = }
                      {           a3 b3 c3           q3 r3 s3      }

                    {a1q1+a2r1+a3s1  b1q1+b2r1+b3s1  c1q1+c2r1+c3s1}
                    {a1q2+a2r2+a3s2  b1q2+b2r2+b3s2  c1q2+c2r2+c3s2}
                    {a1q3+a2r3+a3s3  b1q3+b2r3+b3s3  c1q3+c2r3+c3s3}
BEGIN
  For Col := 1 to MatSize do
  begin
   for Row := 1 to MatSize do
    begin
      Sum := 0;
      for k := 1 to MatSize do             {Calculate:     }
       Sum := Sum + (A[k,Col])*(B[Row,k]); {i.e, a1q1+a2r2+a3s1}
      C[Row,Col] := Sum;
    end; {end Row}
  end;  {end Col}
END;  {end Mult}
{===============================================================}
{This procedure creates a scaling matrix C, x1  0  0
                                            0  y1  0
                                            0  0  z1
such that when multiplied with A, say, we get:

x1  0  0        a1  b1  c1      x1*a1  y1*b1  z1*c1
0  y1  0   x    a2  b2  c2  =   x1*a2  y1*b2  z1*c2
0  0  z1        a3  b3  c3      x1*a3  y1*b3  z1*c3
}
PROCEDURE Scale(X1,Y1,Z1 : integer; Var C : MatType);
VAR
    i,j : word;
BEGIN
    For i := 1 to MatSize do
       For j := 1 to MatSize do
          C[i,j] := 0;
    C[1,1] := X1;
    C[2,2] := Y1;
    C[3,3] := Z1;
END; {end Scale}
{===============================================================}
{This procedure will rotate a point (x,y,z) to (x',y',z') through
 Theta degrees along the M axis and return the transformed matrix
 in C.

 M = 1 for X_axis
 M = 2 for Y_axis
 M = 3 for Z_axis

Rotation about X_axis:
                       1      0      0
                  C =  0      Cos(T) Sin(T)
                       0     -Sin(T) Cos(T)
Rotation about Y_axis:
                       Cos(T) 0     -Sin(T)
                  C =  0      1      0
                       Sin(T) 0      Cos(T)
Rotation about Z_axis:
                       Cos(T) Sin(T) 0
                  C = -Sin(T) Cos(T) 0
                       0      0      1
}
PROCEDURE Rot(M : word; Theta : Real;Var C : MatType);
VAR
   i,j   : word;
   m1,m2 : word;
   Ct,St : real;
BEGIN
  For i := 1 to MatSize do  {set all elements to zero}
    for j := 1 to MatSize do
      C[i,j] := 0;
  C[m,m] := 1;              {Rotation is about this axis and}
  M1 := M  Mod 3 + 1;       {the value does not change.  M1 }
  M2 := M1 Mod 3 + 1;       {and M2 determine where the Sines}
                            {and Cosines go based on M.}
  Ct := Cos(Theta);         {Compute transforming elements}
  St := Sin(Theta);

  C[m1,m1] := Ct;           {Set location of transforming elements}
  C[m2,m2] := Ct;
  C[m1,m2] := St;
  C[m2,m1] := -St;
END; {end Rot}
{===============================================================}
{This procedure will plot (x',y',z').}

PROCEDURE Draw_Solid;
VAR
    j : word;    {Line number}
BEGIN
    For j := 1 to NoLines do
     Make_line(xp[L[1,j]],yp[L[1,j]],xp[L[2,j]],yp[L[2,j]]);
END; {end Draw_Solid}

{This procedure is given the final transform matrix, P, which
 it uses to scale and rotate the (x,y,z) coordinates in the
 X,Y,Z arrays. The points are translated simply by adding the
 translation values to the re-scaled and rotated points.}
{===============================================================}
PROCEDURE Make_Cube(P : MatType);
VAR
    i : word;         {vertex}
                            {Transform (x,y,z) to (x',y',z') and}
BEGIN                       {store in the arrays Xp,Yp, and Zp}
 For i := 1 to NoVerts do
   begin
     Xp[i] := p[1,1]*x[i] + p[1,2]*y[i] + p[1,3]*z[i] + x1;
     Yp[i] := p[2,1]*x[i] + p[2,2]*y[i] + p[2,3]*z[i] + y1;
     Zp[i] := p[3,1]*x[i] + p[3,2]*y[i] + p[3,3]*z[i] + z1;
     Xp[i] := Xp[i]*dist/(Zp[i] + dist);  {Add perspective}
     Yp[i] := Yp[i]*dist/(Zp[i] + dist);  {with distance the object}
   end;                                   {is viewed from}
   Draw_Solid;                            {Draw object}
END; {end Make_cube}
{===============================================================}
PROCEDURE GetInfo_and_set(Var C : MatType);
VAR
    i     : word;
    Angle : integer;
    axis     : array[1..3] of word;  {axis of initial rotation}
    theta    : array[1..3] of real;
{-----------------------------------------------------------}
PROCEDURE Set_Position(Var SetRotMat : MatType);
VAR
    R1,R2,R3,TempMat : MatType;{  This routine sets initial}
BEGIN                          {position of the cube}
  Rot(axis[1],theta[1],R1);    {before it is rotated 360 degs}
  Rot(axis[2],theta[2],R2);    {about one of the axis.}
  Mult(R1,R2,TempMat);         {  The order of multiplication is}
  Rot(axis[3],theta[3],R3);    {important! AxB means rotate with}
  Mult(TempMat,R3,SetRotMat);  {matrix A first then B.}
END; {end Set_Position}
{-----------------------------------------------------------}

BEGIN  {GetInfo_and_set}
  For i := 1 to 3 do       {Get axis and rotation about them}
  begin                    {The order of rotation is significant}
    Write('Enter the axis of rotation: ');
    Readln(axis[i]);
    Write('Enter rotation for axis ',axis[i],': ');
    readln(angle);
    theta[i] := (angle/180)*pi;
  end;
                        {Now get translation points for}
                        {the procedure Make_Cube}
  Write('Enter translation coordinates (x,y,z): ');
  Readln(x1,y1,z1);     {Get scaling factors}
  Write('Enter scaling (x,y,z): ');
  Readln(sx,sy,sz);
  Write('Enter distance solid is viewed from: ');
  Readln(dist);         {Get perspective distance...this must be
                         greater than length of longest axis so
                         observe is outside of object}
  Write('Enter axis to rotate in real time about: ');
  readln(axis_o);       {Get axis to rotate 360 degs about}
  Write('Rotate about Prime (0) or original axis (1): ');
  Readln(Choice);       {Are we rotating about original axis or }
                        {a transformed (prime) axis?}
  Set_Position(C);      {calculate initial rotation of axis}
END; {end GetInfo_and_set}
{===============================================================}
{This procedure will transform P to yield a rotatation
 about axis Axis_o}

PROCEDURE Rot_It(P : MatType);
VAR
    R1,R2,LastRot,FinalMat,ScaleMat : MatType;
    IncAng : word;
    RadAng : real;
BEGIN

  for IncAng := 0 to 60 do      {Go from 0 to 360 degrees}
  begin
    RadAng := IncAng*pi/30;     {using radians}
    rot(axis_o,RadAng,R1);      {calculate rotation matrix, R1}
    if choice = 1 then
       mult(p,R1,LastRot)       {Rotate about original axis}
    else
       mult(R1,p,LastRot);      {Rotate about prime axis}
    Scale(sx,sy,sz,ScaleMat);   {Scale last}
    Mult(ScaleMat,LastRot,FinalMat);
    Setcolor(1);           {Draw}
    Make_Cube(FinalMat);   {Generate solid}
    setcolor(0);           {Erase}
    Make_cube(FinalMat);
  end;
  setcolor(1);
  make_cube(FinalMat);
END; {end Rot_it}
{***************************************************************}
BEGIN                       {Start of Main Body}
  Clrscr;
  Set_up;                   {set up coordinate of cube}
  GetInfo_and_set(P);       {get initial position of cube}
  Gd := Detect;
  InitGraph(Gd, Gm, '');
  if GraphResult <> grOk then
    Halt(1);
  NoVerts := 8;              {cube has 8 corners and 12 lines}
  NoLines := 12;
  Rot_It(P);                 {rotate 360 degs about an axis}
  ch := readkey;
  CloseGraph;
END.
