#include "splay.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

int char_compare(const splay_tree_node *item1,const splay_tree_node *item2);

make_splay_tree_subclass(char, char_compare);

int
char_compare(const splay_tree_node *item1,const splay_tree_node *item2) {
	if(((char_splay_tree_node *)item1)->contained_data < 
		((char_splay_tree_node *)item2)->contained_data)
		return -1;
	else if(((char_splay_tree_node *)item1)->contained_data >
		((char_splay_tree_node *)item2)->contained_data)
		return 1;
	else
		return 0;
}

//
// we need to derive further subclasses in order to allow printing
// out of the tree in our desired format.  The first subclass allows
// tree_scan access to splay tree node internals, and the second adds
// a print method to our char_splay_tree_root

class dirty_trick : public splay_tree_node {
public:
	friend void tree_scan(const dirty_trick *current,
		const int row,const int col,const int interval);
};

class char_splay_tree_rootX : public char_splay_tree_root {
public:
	void print(void);
};

static char dispbuf[32][64];

void tree_scan(const dirty_trick *current,
	const int row,const int col,const int interval) {
	dispbuf[row][col] = ((char_splay_tree_node *)current)->contained_data;
	if(interval > 1) {
		if(current->left != 0)
			tree_scan((dirty_trick *)current->left,
				row+1,col-(interval/2),interval/2);
		if(current->right != 0)
			tree_scan((dirty_trick *)current->right,
				row+1,col+(interval/2),interval/2);
	}
}

void
char_splay_tree_rootX::print(void) {

	if(root == 0)
		return;

	memset(dispbuf,' ',sizeof(dispbuf));
	tree_scan((dirty_trick *)root,0,31,32);

	for(int i = 0; i < 6; i++) {
		for(int j = 0; j < 64; j++)
			putchar(dispbuf[i][j]);
		putchar('\n');
	}
}	

void
build_random_tree(char_splay_tree_root &x) {
	char buf[128];
	for(int i = 0; i < 26; i++) {
		x.insert(((rand() % 26) + 'a'));
	}
}

void printchar(splay_tree_node *item) {
	putchar(((char_splay_tree_node *)item)->contained_data);
}

int
main() {
	char buf[128];
	char_splay_tree_rootX x;
	char *ch;
	// clock is a nonstandard function returning the current
	// MSDOS tick count.  On unix srand(getpid()) works better.
	srand(clock());
	for(int done = 0; !done; ) {
		fputs(
		"[i]nsert,[d]elete,[l]ookup,[p]rint,[r]andom,[t]raverse,[q]uit : ",
		stdout);
		buf[0] = '\0';
		if(gets(buf) == NULL)
			break;
		switch(buf[0]) {
		case 'i':
			x.insert(buf[2]);
			break;
		case 'd':
			x.remove(buf[2]);
			break;
		case 'l':
			ch = x.access(buf[2]);
			printf("%c %s\n",buf[2],
				(ch == 0 ? "not found" : "found"));
			break;
		case 'p':
			x.print();
			break;
		case 'r':
			build_random_tree(x);
			break;
		case 't':
			x.traverse(printchar);
			putchar('\n');
			break;
		case 'q':
			done = 1;
			break;
		}
	}
}
